"""
Pynedra

This module provides a set of functions to communicate with the
webservice of the long-term storage and archiving solution of Synedra.

The module requires valid certificate and certificate key *.pem files
to establish a connection to the server.
"""
import io
from dataclasses import dataclass, field
from typing import Optional, Tuple
import xmltodict
from pydicom import dcmread
from pydicom.dataset import Dataset
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from requests import Session
import xml.etree.ElementTree as ET

from .options import Options, Settings
from .uri import URI, URISuffix


class AlreadyConnected(Exception):
    pass


class CannotConnectError(Exception):
    pass


@dataclass
class Pynedra:
    """
    Class for Synedra Webservice
    """
    uid_tuple: tuple
    base_url: str
    certificate_bundle: Optional[Tuple[str, str]]
    session: Session() = field(init=False, default=None)
    url: URI = field(init=False)

    def __post_init__(self):
        self._start_session()
        self.url = URI(f'{self.base_url}',
                       self.uid_tuple[0],
                       self.uid_tuple[1],
                       self.uid_tuple[2],
                       URISuffix.XML)

    def _start_session(self):
        """
        Custom adapter with timeout properties / options - see implementation in class below

        Setup session and mount the adapter to both the http:// and https:// endpoint
        """
        if self.session is None:
            self.session = Session()
            self.session.cert = self.certificate_bundle
            self.session.verify = False       # should be True!!!
            self.session.stream = False

            try:
                if Settings.total_retries > 0:
                    adapter = HTTPAdapter(
                        max_retries=Retry(total=Settings.total_retries,
                                          backoff_factor=Settings.backoff_factor,
                                          status_forcelist=[413, 429, 500, 502, 503, 504],
                                          allowed_methods=["HEAD", "GET", "OPTIONS"],
                                          ),
                    )
                    for prefix in "http://", "https://":
                        self.session.mount(prefix, adapter)
            except CannotConnectError as ex:
                raise CannotConnectError('Connection not possible!') from ex
        else:
            raise AlreadyConnected('Transport is already connected')

    def get_metadata(self) -> dict:
        """
        Get meta data of dicom image as dictionary.
        """
        with self.session.get(str(self.url)) as results:
            if results.status_code == 200:
                dict_result = xmltodict.parse(results.text)
                return dict_result
            return {}

    def get_dicom(self) -> Optional[Dataset]:
        """
        Get dicom image.
        """
        dicom_content = None
        with self.session.get(str(self.get_url())) as results:
            if results.status_code == 200:
                for chunk in results.iter_content(chunk_size=None, decode_unicode=True):
                    if chunk:
                        dicom_content = dcmread(io.BytesIO(chunk), force=True)
            return dicom_content

    def get_url(self) -> URI:
        """
        Get URI out of different options.
        """
        aim_id = self.get_metadata()['aimInfo']['dicomStudy']['dicomSeries']['dicomImage']['@aimId']
        if Options.xml_header:
            if Options.no_pixel_data and Options.dump:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.XML_NO_PIXEL_DUMP)
            if Options.no_pixel_data:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.XML_NO_PIXEL)
            if Options.dump:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.XML_DUMP)
            return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.XML)

        else:
            if Options.no_pixel_data and Options.dump:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.NO_PIXEL_DUMP)
            if Options.no_pixel_data:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.NO_PIXEL)
            if Options.dump:
                return URI(self.base_url, aim_id=aim_id, suffix=URISuffix.DUMP)
            return URI(self.base_url, aim_id=aim_id)

    def get_xml_file(self) -> ET:
        """
        Get xml header.
        """
        with self.session.get(str(self.get_url())) as results:
            if results.status_code == 200:
                start = results.text.find('<?xml')
                end = results.text.find('</textarea>')
                return ET.fromstring(results.text[start:end])

    def get_file(self):
        """
        Get a file - implementation not yet finished.
        """
        with self.session.get(str(self.get_url())) as results:
            if results.status_code == 200:
                return results



"""
Options with default values
"""
from dataclasses import dataclass, field


@dataclass
class Options:
    """
    Can be changed with e.g. Options.aim_id = False
    """
    xml_header: bool = field(default=False)
    dump: bool = field(default=False)
    no_pixel_data: bool = field(default=False)


@dataclass
class Settings:
    """
    Can be changed with e.g. Settings.total_retries = 5
    """
    total_retries: int = field(default=10)
    timeout_seconds: int = field(default=2)
    backoff_factor: int = field(default=5)
    file_type: str = field(default='image')



"""
Utilities for DICOMweb URI manipulation.
"""
from __future__ import annotations

from enum import Enum
from typing import Optional
from dataclasses import dataclass

from .options import Settings


class URISuffix(Enum):
    XML = 'xml'
    XML_DUMP = 'xml?dump=1'
    XML_NO_PIXEL = 'xml?no_pixel_data=1'
    XML_NO_PIXEL_DUMP = 'xml?no_pixel_data=1&dump=1'
    DUMP = '?dump=1'
    NO_PIXEL = '?no_pixel_data=1'
    NO_PIXEL_DUMP = '?no_pixel_data=1&dump=1'


@dataclass
class URI:
    base_url: str
    study_uid: Optional[str] = None
    series_uid: Optional[str] = None
    sop_uid: Optional[str] = None
    suffix: Optional[URISuffix] = None
    aim_id: Optional[str] = None

    def __post_init__(self):
        self.root_url = self.base_url.split('/refptr')[0]

    def __str__(self) -> str:
        """
        Returns the object as a DICOMweb URI string.
        """
        if self.aim_id is None:
            parts = (self.study_uid, self.series_uid, self.sop_uid)
            dicomweb_suffix = ''.join(f'{part}-' for part in parts if part is not None).rstrip('-')

            # URI for metadata
            if self.suffix is not None:
                dicomweb_suffix = f"{dicomweb_suffix}.{self.suffix.value}"

            return f'{self.base_url}{dicomweb_suffix}'

        else:
            # URI for image extraction
            if self.suffix is not None:
                return f'{self.root_url}/{Settings.file_type}/{self.aim_id}.{self.suffix.value}'

            return f'{self.root_url}/{Settings.file_type}/{self.aim_id}'



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pynedra']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'dicomgenerator>=0.4.0,<0.5.0',
 'factory-boy>=3.2.1,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pydicom>=2.3.0,<3.0.0',
 'pytest>=7.1.2,<8.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'requests>=2.27.1,<3.0.0',
 'setuptools>=62.3.2,<63.0.0',
 'urllib3>=1.26.9,<2.0.0',
 'xmltodict>=0.13.0,<0.14.0']

setup_kwargs = {
    'name': 'pynedra',
    'version': '0.1.5',
    'description': 'Pynedra',
    'long_description': 'None',
    'author': 'barbara73',
    'author_email': 'barbara.jesacher@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
